import scala.annotation.tailrec

package object main {
  val foo = 42
  val bar = 2 + foo
  
  def quadrado(x: Double) = x * x
  def somaDeQuadrados(x: Double, y: Double) =
    quadrado(x) + quadrado(y)
  
  def loop: Double = loop
  def primeiro(x: Double, y: => Double) = x
  
  def abs(x: Double) = if (x >= 0) x else -x
  
  def raiz(x: Double): Double = {
    val erro = 0.0000001
    
    @tailrec
    def raizIter(est: Double, x: Double): Double =
      if (suficiente(est, x)) est else raizIter(melhora(est, x), x)

    def suficiente(est: Double, x: Double) = 
      abs(quadrado(est) - x) < erro

    def melhora(est: Double, x: Double) = (est + x / est) / 2
     
    raizIter(1.0, x)
  }
  
  val x = 0
  def f(y: Int) = y + 1  
  val result = {
    val x = f(3)
    x * x
  }
  
  @tailrec
  def mdc(a: Int, b: Int): Int = 
    if (b == 0) a else mdc(b, a % b)
  
  @tailrec
  def contador(a: Double, b: Double): Double =
    if (a == 0) a else contador(a - b, b)
  
  def fat(x: Int): Int = if (x < 2) 1 else x * fat(x - 1)

  def fatIter(x: Int) = {
    @tailrec
    def loop(acum: Int, x: Int): Int =
      if (x < 2) acum else loop(acum * x, x - 1)
    
    loop(1, x)
  }
  
  val l = List(1, 2) :: List(3) :: List(4, 5) :: Nil
  
  @tailrec
  def elem[T](i: Int, l: List[T]): T = 
    if (i < 0) error("índice não pode ser negativo")
    else l match {
      case Nil => error("índice fora dos limites da lista")
      case h :: t => if (i == 0) h else elem(i-1, t)
    }
        
  def tamanho[T](l: List[T]): Int = l match {
    case Nil => 0
    case _ :: t => 1 + tamanho(t)
  }
  
 def tamanhoRF[T](l: List[T]): Int = {
    @tailrec
    def loop(acum: Int, l: List[T]): Int = l match {
     	case Nil => acum
        case _ :: t => loop(acum+1, t)
    }
    loop(0, l)  
  }
    
  def concat[T](l1: List[T], l2: List[T]): List[T] = (l1, l2) match {
    case (Nil, l) => l
    case (l, Nil) => l
    case (h :: t, l) => h :: concat(t, l)
  }
  
  def inverte[T](l: List[T]): List[T] = {
    @tailrec
    def loop(acum: List[T], l: List[T]): List[T] =
      if (l.isEmpty) acum
      else loop(l.head :: acum, l.tail)
    loop(Nil, l)
  }
    
  def concatRF[T](l1: List[T], l2: List[T]): List[T] = {
    @tailrec
    def loop(acum: List[T], l: List[T]): List[T] =
      if (l.isEmpty) acum
      else loop(l.head :: acum, l.tail)
    
    inverte(loop(inverte(l1), l2))
  }
  
  // Zipper
  
  def zipper[T](l: List[T]): (List[T], T, List[T]) = l match {
    case Nil => error("lista vazia não tem zipper")
    case h :: t => (Nil, h, t)
  }
  
  def avance[T](z: (List[T], T, List[T])): (List[T], T, List[T]) =
    z match {
      case (_, _, Nil) => error("estou no final da lista")
      case (e, f, h :: t) => (f :: e, h, t)
    }

  def retroceda[T](z: (List[T], T, List[T])): (List[T], T, List[T]) =
    z match {
      case (Nil, _, _) => error("estou no inicio da lista")
      case (h :: t, f, d) => (t, h, f :: d)
    }
  
  // Ambientes
  
  def lookup[T](env: List[(String, T)], nome: String): T = env match {
    case Nil => error(nome + " não encontrado")
    case (n, e) :: t => if (nome == n) e else lookup(t, nome) 
  }
  
  def vazio[T]: List[(String, T)] = Nil
  
  def extend[T](env: List[(String, T)], n: String, v: T) = (n, v) :: env
  
  // Somatórios
  
  def somaInt(a: Int, b: Int): Int = if (a > b) 0 else a + somaInt(a + 1, b)


  def somaQuad(a: Int, b: Int): Int = {
    def quadrado(x: Int) = x * x
  
    if (a > b) 0 else quadrado(a) + somaQuad(a + 1, b)
  }
  
  def soma(f: Int => Int, a: Int, b: Int): Int = 
    if (a > b) 0 else f(a) + soma(f, a + 1, b)
    
  def produto(f: Int => Int, a: Int, b: Int): Int = 
    if (a > b) 1 else f(a) * produto(f, a + 1, b)

  def fatProd(n: Int) = produto(x => x, 2, n)
  
  def msoma(f: Int => Int, z: Int, op: (Int, Int) => Int, a: Int, b: Int): Int =
    if (a > b) z else op(f(a), msoma(f, z, op, a+1, b))
 
  def soma1(f: Int => Int, a: Int, b: Int) = msoma(f, 0, (x, y) => x + y, a, b)

  def produto1(f: Int => Int, a: Int, b: Int) = msoma(f, 1, (x, y) => x * y, a, b)
  
  def pontoFixo(f: Double => Double)(est: Double): Double = {
    val erro = 0.000001
    def suficiente(est1: Double, est2: Double) =
      abs(est2 - est1) < erro
    @tailrec
    def loop(est: Double): Double = {
      val prox = f(est)
      if (suficiente(est, prox)) prox
      else loop(prox)
    }
    loop(est)
  }
  
  def derivada(f: Double => Double)(x: Double) = {
    val dx = 0.000001
    (f(x + dx) - f(x)) / dx
  }
  
  def newton(g: Double => Double): Double =
    pontoFixo(x => x - g(x) / derivada(g)(x))(1.0)

  def raizpf(x: Double): Double = pontoFixo(y => (y + x/y)/2)(1.0)
  
  def raiznewton(x: Double): Double = newton(y => y*y - x)
  
  def map[T,U](f: T => U)(l: List[T]): List[U] = l match {
    case Nil => Nil
    case h :: t => f(h) :: map(f)(t)
  }
  
  def multLista(l: List[Double], fator: Double): List[Double] =
    map((x: Double) => x * fator)(l)
    
  def filter[T](pred: T => Boolean)(l: List[T]): List[T] = l match {
    case Nil => Nil
    case h :: t => if (pred(h)) h :: filter(pred)(t)
        else filter(pred)(t)
  }
  
  def filtraMaiores(l: List[Double], pivo: Double): List[Double] =
    filter((x: Double) => x > pivo)(l)

  /*
   * soma(List(1,2,3)) =
   * 1 + soma(List(2,3)) =
   * 1 + (2 + soma(List(3))) =
   * 1 + (2 + 3)  
   */
   
  def reduceRight[T](op: (T, T) => T)(l: List[T]): T = l match {
    case Nil => error("lista vazia")
    case h :: Nil => h
    case h :: t => op(h, reduceRight(op)(t))
  }
  
  def reduceLeft[T](op: (T, T) => T)(l: List[T]): T = {
    @tailrec
    def loop(l: List[T], acum: T): T = l match {
      case Nil => acum
      case h :: t => loop(t, op(h, acum))
    }
    l match {
      case Nil => error("lista vazia")
      case h :: t => loop(t, h)
    }
  }
  
  def foldRight[T,U](op: (T, U) => U, z: U)(l: List[T]): U = l match {
    case Nil => z
    case h :: t => op(h, foldRight(op, z)(t))
  }
  
  @tailrec
  def foldLeft[T,U](op: (U, T) => U, z: U)(l: List[T]): U = l match {
    case Nil => z
    case h :: t => foldLeft(op, op(z, h))(t)
  }

  def lookupFold[T](env: List[(String, T)], nome: String): Option[T] = 
    foldLeft((res: Option[T], elem: (String, T)) => res match {
                                                      case Some(_) => res
                                                      case None => {
                                                        val (n, e) = elem
                                                        if (n == nome)
                                                          Some(e)
                                                        else
                                                          None 
                                                      }
                                                    }, None)(env)
  
  
  /* l1_1 op (l1_2 op (l1_3 ... (l1_n op z) ... )) */                                                  
  def concatFold[T](l1: List[T], l2: List[T]): List[T] =
    foldRight((h: T, t: List[T]) => h :: t, l2)(l1)
  
  /* ( ... (z op l_1) op l_2) ... op l_n =>
   *     l_n :: ... l_2 :: l_1 :: z
   */
  def inverteFold[T](l: List[T]): List[T] =
    l.foldLeft(List[T]())((t: List[T], h: T) => h :: t)
    //foldLeft((t: List[T], h: T) => h :: t, Nil)(l)
    
  def mapFold[T, U](f: T => U)(l: List[T]): List[U] =
    foldRight((h: T, t: List[U]) => f(h) :: t, Nil)(l)
    
  def prodCart1[A,B](l1: List[A], l2: List[B]): List[(A,B)] =
    l1.flatMap(x => l2.map(y => (x, y)))//.flatten
      //foldRight[List[(A,B)]](Nil)((l1, l2) => l1 ++ l2)
    
  def paresPrimos(l1: List[Int], l2: List[Int]): List[(Int, Int)] =
    prodCart1(l1, l2).filter({ case (x, y) => mdc(x, y) == 1 })
    
  def paresPrimosFor(l1: List[Int], l2: List[Int]): List[(Int, Int)] =
    for {
      x <- l1  // l1 gera x
      y <- l2  // l2 gera y
      if mdc(x, y) == 1 // filtro
    } yield (x, y)   // { (x,y) | x E l1 ^ y E l2 ^ mdc(x,y) = 1 }

  def produtoInt(l1: List[Int], l2: List[Int]): Int =
    (for {
      x <- l1  // l1 gera x
      y <- l2  // l2 gera y
      if mdc(x, y) == 1 // filtro
    } yield (x, y)).foldLeft(0)((x: Int, y: (Int, Int)) => x + y._1 * y._2)

  def produtoInt2(l1: List[Int], l2: List[Int]): Int =
    (for {
      (x, y) <- for {
                  x <- l1  // l1 gera x
                  y <- l2  // l2 gera y
                  if mdc(x, y) == 1 // filtro
                } yield (x, y)
    } yield x * y).sum

  def produtoInt3(l1: List[Int], l2: List[Int]): Int =
    (for {
       x <- l1  // l1 gera x
       y <- l2  // l2 gera y
       if mdc(x, y) == 1 // filtro
     } yield x * y).sum
     
  /*
   for {
      x <- l1  // l1 gera x
      y <- l2  // l2 gera y
      if mdc(x, y) == 1 // filtro
   } yield (x, y)
    
   --passo 1 (caso 2)-->
    
   l1.flatMap({ case x => for {
                            y <- l2
                            if mdc(x, y) == 1
                          } yield (x, y) }) 

   --passo 2 (caso 1)-->

   l1.flatMap({ case x => for {
                            y <- l2.filter({ case y => mdc(x, y) == 1 })
                          } yield (x, y) }) 

   --passo 3 (caso 3)-->

   l1.flatMap({ case x => l2.filter({ case y => mdc(x, y) == 1 }).map(
                                 { case y => (x, y) }) }) 

   --simplificação-->
   
   l1.flatMap(x => l2.filter(y => mdc(x, y) == 1).map(y => (x, y))) 
   
    
  */
     
  trait ListaInt {
    def soma: Int = this match {
      case NilInt() => 0
      case ConsInt(h, t) => h + somaListaInt(t)
    }
    
    def primeiro: Opcional[Int] = this match {
      case NilInt() => Nada()
      case ConsInt(h, t) => Algum(h)
    }
  }
  case class NilInt() extends ListaInt
  case class ConsInt(h: Int, t: ListaInt) extends ListaInt
  
  def somaListaInt(l: ListaInt): Int = l match {
      case NilInt() => 0
      case ConsInt(h, t) => h + somaListaInt(t)
  }
  
  trait Opcional[T] {
    def map[U](f: T => U): Opcional[U] = this match {
      case Nada() => Nada()
      case Algum(x) => Algum(f(x))
    }
    def flatMap[U](f: T => Opcional[U]): Opcional[U] = this match {
      case Nada() => Nada()
      case Algum(x) => f(x)
    }
    def filter(f: T => Boolean) = this match {
      case Nada() => Nada()
      case Algum(x) => if(f(x)) Algum(x) else Nada() 
    }
  }
  case class Nada[T]() extends Opcional[T]
  case class Algum[T](x: T) extends Opcional[T]
  
  def somaPrimeiros(l1: ListaInt, l2: ListaInt): Opcional[Int] = for {
    x <- l1.primeiro
    y <- l2.primeiro
  } yield x + y  /* ---> 
  * 
  * l1.primeiro.flatMap(x => l2.primeiro.map(y => x + y))
  * 
  */
}


    
